/*
 * spr2pc.c: Convert the .SPR files on the Final Fantasy Tactics
 *	CD into PC images using the GD library. This is not
 *	intended for copyright infringement.
 *
 * Copyright 2000, Chris Davenport, George Greer, Corey Sanders
 *	<davenpcj@muohio.edu>
 *	<greerga@muc.muohio.edu>
 *	<csanders@muc.muohio.edu>
 *
 * Licensed under the GNU General Public License (GPL)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *   or visit ftp://ftp.gnu.org/pub/gnu/COPYING via the Internet.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <assert.h>

#include <getopt.h>
#include <gd.h>
#include <sys/param.h>

#include "mmaplib.h"

#define FALSE	0
#define TRUE	(!FALSE)

static const char *copyright = "spr2pc.c: Copyright 2000, MU FFT Team\n"
         "Chris Davenport <davenpcj@muohio.edu>,\n"
         "George Greer <greerga@muc.muohio.edu>,\n"
         "Corey Sanders <csanders@muc.muohio.edu>";

static const char *version = "v0.5 (Build: "__DATE__" "__TIME__")";

/* --------------------------- */

enum image_e {
  IMG_JPEG = 0,
  IMG_PNG = 1,
  IMG_GD = 2,
};

struct options_s {
  unsigned int all : 1;
  unsigned int head : 1;
  unsigned int color : 1;
  unsigned char uniform;
  unsigned int xscale;
  unsigned int yscale;
  unsigned int verbose;
  unsigned int blocklen;
  long skipdata;
  int headx;
  int heady;
  size_t paletteoffset;
  enum image_e format;
  char *inputpalette;
  char *outputpalette;
  char *outputfile;
};

/*
 * The Meat.
 * --------------------------------------------------------------------------------------------------------------------------------
 */

#define HEADER_SIZE	768
#define BPP		2
#define PALETTE_BYTES	32
#define PALETTE_SIZE	(PALETTE_BYTES / BPP)

union palette_u {
  struct {
    unsigned int r : 5;
    unsigned int g : 5;
    unsigned int b : 5;
    unsigned int t : 1;
  } bits;
  unsigned short int bytes;
};
	
/*
 * 16-bit color, 5551, Red-Green-Blue-Transparent
 */
struct palette_s {
  unsigned char r;
  unsigned char g;
  unsigned char b;
  unsigned char t : 1;
};

int make_bw_palette(struct palette_s **palret, int body)
{
  int offset = body ? 0 : 255;
  int itr;
  struct palette_s *pal;

  pal = *palret = (struct palette_s *)calloc(256, sizeof(struct palette_s));
  for (itr = 0; itr < 256; itr++)
    pal[itr].r = pal[itr].g = pal[itr].b = offset - itr;

  return 256;
}

int make_color_palette(struct palette_s **palret, const unsigned char *data, struct options_s *opts, int body)
{
  int itr, spare, total = 0;
  struct palette_s *pal = *palret = (struct palette_s *)calloc(PALETTE_SIZE, sizeof(struct palette_s));

  if (strncmp(data, "GIMP Palette", 12) != 0) {
    /* SPR Palette */
    int uniform = opts->uniform * PALETTE_BYTES;
    int offset = body ? 0 : 4 * PALETTE_BYTES;
    union palette_u convpal;

    for (itr = 0; itr < PALETTE_SIZE; itr++) {
      memcpy(&convpal.bytes, data + (itr + uniform + offset) * BPP, BPP);
      pal[itr].r = convpal.bits.r * 255 / 0x1F;
      pal[itr].g = convpal.bits.g * 255 / 0x1F;
      pal[itr].b = convpal.bits.b * 255 / 0x1F;
    }
  } else {
    total = strcspn(data, "\n");
    /* GIMP Palette */
    for (itr = 0; itr < PALETTE_SIZE; itr++) {
      unsigned int r, g, b;
      sscanf(data + total, " %d %d %d %n", &r, &g, &b, &spare);
      pal[itr].r = r;
      pal[itr].g = g;
      pal[itr].b = b;
      total += spare;
    }
  }

  return PALETTE_SIZE;
}

/*
 * Assumes HEADER_SIZE-bytes available for reading.
 */
int make_palette(struct palette_s **palret, const unsigned char *data, struct options_s *opts, int body)
{
  if (opts->color)
    return make_color_palette(palret, data + opts->paletteoffset, opts, body);
  else
    return make_bw_palette(palret, body);
}

int make_image(gdImagePtr img, struct palette_s *pal, struct palette_s *head, size_t palsize, const unsigned char *data, size_t length, struct options_s *options)
{
  size_t pos;
  union palette_u convpal;
  struct palette_s *thispal;
  int x1, y1, x2, y2, cr, ix;

  for (pos = 0; pos < length; pos++) {
    x1 = (pos % options->blocklen) * options->xscale;
    y1 = (pos / options->blocklen) * options->yscale;
    x2 = x1 + (options->xscale - 1);
    y2 = y1 + (options->yscale - 1);

    ix = data[pos] % palsize;

    if (options->head &&
        y1 >= (options->heady + options->all * 6) && y1 < (options->heady + 32 * options->yscale + options->all * 6) &&
        x1 >= (options->headx) && x1 < (options->headx + 24 * options->xscale))
          thispal = &head[ix];
    else
      thispal = &pal[ix];

    convpal.bits.r = thispal->r;
    convpal.bits.g = thispal->g;
    convpal.bits.b = thispal->b;
    convpal.bits.t = thispal->t;

    if ((cr = gdImageColorExact(img, thispal->r, thispal->g, thispal->b)) < 0)
      if ((cr = gdImageColorAllocate(img, thispal->r, thispal->g, thispal->b)) < 0)
        error("ran out of colors");

    gdImageFilledRectangle(img, x1, y1, x2, y2, cr);
  }

  return 0;
}

int write_palette(char *palettefile, struct palette_s *body_palette, struct palette_s *head_palette, size_t palsize, struct options_s *options)
{
  FILE *out;
  int itr;
  char filename[PATH_MAX];

  snprintf(filename, sizeof(filename), "%s.body", palettefile);
  out = fopen(filename, "w");
  fputs("GIMP Palette\n", out);
  for (itr = 0; itr < palsize; itr++)
    fprintf(out, "%3d %3d %3d\n", body_palette[itr].r, body_palette[itr].g, body_palette[itr].b);
  fclose(out);

  if (!head_palette || !options->head)
    return 0;

  snprintf(filename, sizeof(filename), "%s.head", palettefile);
  out = fopen(filename, "w");
  fputs("GIMP Palette\n", out);
  for (itr = 0; itr < palsize; itr++)
    fprintf(out, "%3d %3d %3d\n", head_palette[itr].r, head_palette[itr].g, head_palette[itr].b);
  fclose(out);

  return 0;
}

int spr2pc(const unsigned char *data, size_t length, void *extra)
{
  /* Palette and file variables. */
  struct options_s *options = (struct options_s *)extra;
  struct palette_s *body_palette = NULL, *head_palette = NULL;
  size_t palsize;
  FILE *outfile;

  /* Image variables */
  gdImagePtr img;
  size_t xsize, ysize;
  long start;

  if (!(outfile = fopen(options->outputfile, "wb")))
    error("open output file");

  if (!options->all && !options->skipdata)
    start = HEADER_SIZE;
  else
    start = (options->skipdata >= 0 ? options->skipdata : length + options->skipdata);

  xsize = options->blocklen * options->xscale;
  ysize = (options->all ? (length - start) / options->blocklen : 286) * options->yscale;

  if (options->verbose)
    printf("Length: %d\nStart : %ld\nX Size: %d\nY Size: %d\n", length, start, xsize, ysize);

  if (!options->inputpalette || !options->color) {
    palsize = make_palette(&body_palette, data, options, TRUE);
    if (options->head)
      make_palette(&head_palette, data, options, FALSE);
  } else {
    void *map;
    char *head;

    if ((head = strchr(options->inputpalette, ':')))
      *head++ = '\0';

    if (!(map = mapopen(options->inputpalette)))
      error(options->inputpalette);
    palsize = make_palette(&body_palette, mapdata(map), options, TRUE);
    mapclose(map);

    if (options->head) {
      if (!head)
        options->head = FALSE;
      else {
        if (!(map = mapopen(head)))
          error(head);
        make_palette(&head_palette, mapdata(map), options, FALSE);
        mapclose(map);
      }
    }
  }
    
  img = gdImageCreate(xsize, ysize);
  gdImageColorTransparent(img, -1);

  make_image(img, body_palette, head_palette, palsize, data + start, length - start, options);

#ifdef GD_GIF
  gdImageGif(img, outfile);
#else
  switch (options->format) {
    case IMG_JPEG:
      gdImageJpeg(img, outfile, -1);  /* -1, default quality */
      break;
    case IMG_PNG:
      gdImagePng(img, outfile);
      break;
    case IMG_GD:
      gdImageGd(img, outfile);
      break;
  }
#endif
  fclose(outfile);

  gdImageDestroy(img);

  if (options->outputpalette)
    write_palette(options->outputpalette, body_palette, head_palette, palsize, options);

  return 0;
}

/*
 * --------------------------------------------------------------------------------------------------------------------------------
 * End of the meat, start option parsing and file handing.
 */

int print_version(const char *progname)
{
  printf("spr2pc.c: Convert FFT SPR files to PC formats.\n\n%s\n\n%s: version %s\n", copyright, progname, version);
  return 0;
}

int usage(const char *progname)
{
  printf("%s [-a] [-b] [-c] [-h] [-u number] [-o outfilename] infilename\n"
         "	-a[offset]	Display all of file, even header/footer.\n"
         "			Offset in file to find palette, optional.\n"
         "	-b		Use black & white approximation.\n"
         "	-c		Use color.\n"
         "	-h[x,y]		Attempt to correctly colorize a head.\n"
         "			x,y is coordinate of top left corner, optional.\n"
#ifndef GD_GIF
         "	-f format	Write out image in this format, supported:\n"
         "		jpeg	Lossy JPEG format.\n"
         "		png	Portable Network Graphic. (Netscape/IE 4+)\n"
         "		gd	Internal GD format.\n"
#endif
         "	-j offset	Start rendering image from offset.\n"
         "	-o filename	Filename to write resulting image to.\n"
         "	-p body[:head]	Import palette from a file. (GIMP or SPR)\n"
         "			:filename for second head palette, optional.\n"
         "	-P filename	Export palette to a file in GIMP format.\n"
         "	-u number	Uniform number (1-8) to write image with.\n"
         "	-v		Verbose output.\n"
         "	filename	The file to use as input.\n"
         "\n"
         "%s --version	Version and copyright information.\n"
         "%s --help	This help message.\n", progname, progname, progname);
  return 1;
}

struct option /* {
  const char *name;
  int has_arg;
  int *flag;
  int val;
} */ long_options[] = {
  { "help",	FALSE, NULL, 0 },
  { "version",	FALSE, NULL, 1 },
  { 0, 0, 0, 0 }
};

const char *imgformat[] = { "JPEG", "PNG", "GD" };
const char *imgformatext[] = { "jpg", "png", "gd" };

int main(int argc, char *argv[])
{
  int gto, zoom;
  struct options_s options;

  memset(&options, 0, sizeof(options));

  options.all = FALSE;
  options.color = TRUE;
  options.xscale = 2;
  options.yscale = 1;
  options.headx = -1;
  options.heady = -1;
  options.paletteoffset = 0;
  options.blocklen = 128;
  options.skipdata = 0;
  options.format = IMG_PNG;

  while ((gto = getopt_long(argc, argv, ":a::bcf:h::j:l:o:p:P:u:vz:",long_options, NULL)) >= 0) {
    switch (gto) {
    case 0:
      return usage(argv[0]);
    case 1:
      return print_version(argv[0]);
    case 'a':
      options.all = !options.all;
      if (optarg)
        options.paletteoffset = atoi(optarg);
      break;
    case 'b':
      options.color = FALSE;
      break;
    case 'c':
      options.color = TRUE;
      break;
    case 'h':
      options.head = !options.head;
      if (optarg)
        sscanf(optarg, "%d,%d", &options.headx, &options.heady);
      break;
    case 'f':
#ifdef GD_GIF
      error("GD library only supports GIF.");
#else
      if (!strcasecmp(optarg, "jpeg"))
        options.format = IMG_JPEG;
      else if (!strcasecmp(optarg, "png"))
        options.format = IMG_PNG;
      else if (!strcasecmp(optarg, "gd"))
        options.format = IMG_GD;
      else
        error("invalid image format");
      break;
#endif
    case 'j':
      options.skipdata = atoi(optarg);
      break;
    case 'l':
      options.blocklen = atoi(optarg);
      break;
    case 'o':
      options.outputfile = optarg;
      break;
    case 'p':
      options.inputpalette = optarg;
      break;
    case 'P':
      options.outputpalette = optarg;
      break;
    case 'u':
      options.uniform = atoi(optarg);
      if (options.uniform < 1 || options.uniform > 8) {
        fprintf(stderr, "Uniform %d is not a valid choice. Use 1-8.\n", options.uniform);
        return 1;
      }
      break;
    case 'v':
      options.verbose++;
      break;
    case 'z':
      sscanf(optarg, "%d:%d:%d", &zoom, &options.xscale, &options.yscale);
      options.xscale *= zoom;
      options.yscale *= zoom;
      break;
    case ':':
      fprintf(stderr, "Option -%c requires an extra argument.\n", optopt);
      return 1;
    case '?':
      fprintf(stderr, "Unknown option -%c.\n", optopt);
      return 1;
    default:
      fputs("Unknown error condition.\n", stderr);
      return 1;
    }
  }

  if (argc == optind)
    fputs("Program requires a filename to translate.\nTry '--help' for more information.", stderr);
  else if (!options.color && options.uniform)
    fputs("Uniform rendering option useless without color.\n", stderr);
  else if (options.xscale == 0 || options.yscale == 0)
    fputs("Scaling by 0 is not useful.\n", stderr);
  else {
    void *handle;

    if (options.uniform > 0)
      options.uniform--;

    if (options.headx < 0)
      options.headx = 40 * options.xscale;	/* 40..64 */

    if (options.heady < 0)
      options.heady = 254 * options.yscale;	/* 254..286 */

    if (options.outputfile == NULL) {
      options.outputfile = (char *)malloc(strlen(argv[optind]) + 4);
      sprintf(options.outputfile, "%s.%s", argv[optind], imgformatext[options.format]);
    } else if (strstr(options.outputfile, "{}")) {
      const char *extn = imgformatext[options.format];
      char *ofile = options.outputfile,
           *curly = strstr(options.outputfile, "{}"),
           *nfile = (char *)malloc(strlen(options.outputfile) + strlen(extn));

      *curly = '\0';
      options.outputfile = nfile;
      sprintf(options.outputfile, "%s%s", ofile, extn);
    }


    if (options.verbose) {
      printf("Options:\n"
	"	All data:	%s\n"
	"	Colorize:	%s\n"
	"	Uniform #:	%d\n"
	"	Verbose:	%d\n"
	"\n"
	"	X Scale:	%d\n"
	"	Y Scale:	%d\n"
	"	Render Offset:	%ld\n"
	"	Block Length:	%d\n"
	"\n"
	"	Head Render:	%s\n"
	"	Head X Offset:	%d\n"
	"	Head Y Offset:	%d\n"
	"\n"
	"	Palette Offset:	%d bytes\n"
	"	Input Palette:	[%s]\n"
	"	Output Palette:	[%s]\n"
	"\n"
	"	Input File:	[%s]\n"
	"	Output File:	[%s]\n"
	"	Output Format:	%s\n",
	options.all ? "YES" : "NO",
	options.color ? "YES" : "NO",
	options.uniform + 1, options.verbose,
	options.xscale, options.yscale, options.skipdata, options.blocklen,
	options.head ? "YES" : "NO", options.headx, options.heady,
	options.paletteoffset, options.inputpalette, options.outputpalette,
	argv[optind], options.outputfile, imgformat[options.format]);
    }

    handle = mapopen(argv[optind]);
    spr2pc(mapdata(handle), maplength(handle), &options);
    mapclose(handle);
  }
  return 1;
}
